@@ -21,16 +21,31 @@ Command-line interface to the Neutron APIs
 from __future__ import print_function
 
 import argparse
+import getpass
+import inspect
+import itertools
 import logging
 import os
 import sys
 
+from keystoneclient.auth.identity import v2 as v2_auth
+from keystoneclient.auth.identity import v3 as v3_auth
+from keystoneclient import discover
+from keystoneclient.openstack.common.apiclient import exceptions as ks_exc
+from keystoneclient import session
+from oslo_utils import encodeutils
+import six.moves.urllib.parse as urlparse
+
 from cliff import app
 from cliff import commandmanager
 
 from neutronclient.common import clientmanager
+from neutronclient.common import command as openstack_command
 from neutronclient.common import exceptions as exc
+from neutronclient.common import extension as client_extension
 from neutronclient.common import utils
+from neutronclient.i18n import _
+from neutronclient.neutron.v2_0 import address_scope
 from neutronclient.neutron.v2_0 import agent
 from neutronclient.neutron.v2_0 import agentscheduler
 from neutronclient.neutron.v2_0 import credential
@@ -42,9 +57,13 @@ from neutronclient.neutron.v2_0.fw import firewallrule
 from neutronclient.neutron.v2_0.lb import healthmonitor as lb_healthmonitor
 from neutronclient.neutron.v2_0.lb import member as lb_member
 from neutronclient.neutron.v2_0.lb import pool as lb_pool
+from neutronclient.neutron.v2_0.lb.v2 import healthmonitor as lbaas_healthmon
+from neutronclient.neutron.v2_0.lb.v2 import listener as lbaas_listener
+from neutronclient.neutron.v2_0.lb.v2 import loadbalancer as lbaas_loadbalancer
+from neutronclient.neutron.v2_0.lb.v2 import member as lbaas_member
+from neutronclient.neutron.v2_0.lb.v2 import pool as lbaas_pool
 from neutronclient.neutron.v2_0.lb import vip as lb_vip
 from neutronclient.neutron.v2_0 import metering
-from neutronclient.neutron.v2_0.nec import packetfilter
 from neutronclient.neutron.v2_0 import netpartition
 from neutronclient.neutron.v2_0 import network
 from neutronclient.neutron.v2_0 import networkprofile
@@ -52,17 +71,20 @@ from neutronclient.neutron.v2_0.nsx import networkgateway
 from neutronclient.neutron.v2_0.nsx import qos_queue
 from neutronclient.neutron.v2_0 import policyprofile
 from neutronclient.neutron.v2_0 import port
+from neutronclient.neutron.v2_0.qos import bandwidth_limit_rule
+from neutronclient.neutron.v2_0.qos import policy as qos_policy
+from neutronclient.neutron.v2_0.qos import rule as qos_rule
 from neutronclient.neutron.v2_0 import quota
+from neutronclient.neutron.v2_0 import rbac
 from neutronclient.neutron.v2_0 import router
 from neutronclient.neutron.v2_0 import securitygroup
 from neutronclient.neutron.v2_0 import servicetype
 from neutronclient.neutron.v2_0 import subnet
+from neutronclient.neutron.v2_0 import subnetpool
 from neutronclient.neutron.v2_0.vpn import ikepolicy
 from neutronclient.neutron.v2_0.vpn import ipsec_site_connection
 from neutronclient.neutron.v2_0.vpn import ipsecpolicy
 from neutronclient.neutron.v2_0.vpn import vpnservice
-from neutronclient.openstack.common.gettextutils import _
-from neutronclient.openstack.common import strutils
 from neutronclient.version import __version__
 
 
@@ -79,10 +101,25 @@ def run_command(cmd, cmd_parser, sub_argv):
         _argv = sub_argv[:index]
         values_specs = sub_argv[index:]
     known_args, _values_specs = cmd_parser.parse_known_args(_argv)
+    if(isinstance(cmd, subnet.CreateSubnet) and not known_args.cidr):
+        cidr = get_first_valid_cidr(_values_specs)
+        if cidr:
+            known_args.cidr = cidr
+            _values_specs.remove(cidr)
     cmd.values_specs = (index == -1 and _values_specs or values_specs)
     return cmd.run(known_args)
 
 
+def get_first_valid_cidr(value_specs):
+    # Bug 1442771, argparse does not allow optional positional parameter
+    # to be separated from previous positional parameter.
+    # When cidr was separated from network, the value will not be able
+    # to be parsed into known_args, but saved to _values_specs instead.
+    for value in value_specs:
+        if utils.is_valid_cidr(value):
+            return value
+
+
 def env(*_vars, **kwargs):
     """Search for the first defined of possibly many env vars.
 
@@ -97,7 +134,23 @@ def env(*_vars, **kwargs):
     return kwargs.get('default', '')
 
 
+def check_non_negative_int(value):
+    try:
+        value = int(value)
+    except ValueError:
+        raise argparse.ArgumentTypeError(_("invalid int value: %r") % value)
+    if value < 0:
+        raise argparse.ArgumentTypeError(_("input value %d is negative") %
+                                         value)
+    return value
+
+
+class BashCompletionCommand(openstack_command.OpenStackCommand):
+    """Prints all of the commands and options for bash-completion."""
+    resource = "bash_completion"
+
 COMMAND_V2 = {
+    'bash-completion': BashCompletionCommand,
     'net-list': network.ListNetwork,
     'net-external-list': network.ListExternalNetwork,
     'net-show': network.ShowNetwork,
@@ -109,6 +162,11 @@ COMMAND_V2 = {
     'subnet-create': subnet.CreateSubnet,
     'subnet-delete': subnet.DeleteSubnet,
     'subnet-update': subnet.UpdateSubnet,
+    'subnetpool-list': subnetpool.ListSubnetPool,
+    'subnetpool-show': subnetpool.ShowSubnetPool,
+    'subnetpool-create': subnetpool.CreateSubnetPool,
+    'subnetpool-delete': subnetpool.DeleteSubnetPool,
+    'subnetpool-update': subnetpool.UpdateSubnetPool,
     'port-list': port.ListPort,
     'port-show': port.ShowPort,
     'port-create': port.CreatePort,
@@ -145,6 +203,31 @@ COMMAND_V2 = {
     'security-group-rule-show': securitygroup.ShowSecurityGroupRule,
     'security-group-rule-create': securitygroup.CreateSecurityGroupRule,
     'security-group-rule-delete': securitygroup.DeleteSecurityGroupRule,
+    'lbaas-loadbalancer-list': lbaas_loadbalancer.ListLoadBalancer,
+    'lbaas-loadbalancer-show': lbaas_loadbalancer.ShowLoadBalancer,
+    'lbaas-loadbalancer-create': lbaas_loadbalancer.CreateLoadBalancer,
+    'lbaas-loadbalancer-update': lbaas_loadbalancer.UpdateLoadBalancer,
+    'lbaas-loadbalancer-delete': lbaas_loadbalancer.DeleteLoadBalancer,
+    'lbaas-listener-list': lbaas_listener.ListListener,
+    'lbaas-listener-show': lbaas_listener.ShowListener,
+    'lbaas-listener-create': lbaas_listener.CreateListener,
+    'lbaas-listener-update': lbaas_listener.UpdateListener,
+    'lbaas-listener-delete': lbaas_listener.DeleteListener,
+    'lbaas-pool-list': lbaas_pool.ListPool,
+    'lbaas-pool-show': lbaas_pool.ShowPool,
+    'lbaas-pool-create': lbaas_pool.CreatePool,
+    'lbaas-pool-update': lbaas_pool.UpdatePool,
+    'lbaas-pool-delete': lbaas_pool.DeletePool,
+    'lbaas-healthmonitor-list': lbaas_healthmon.ListHealthMonitor,
+    'lbaas-healthmonitor-show': lbaas_healthmon.ShowHealthMonitor,
+    'lbaas-healthmonitor-create': lbaas_healthmon.CreateHealthMonitor,
+    'lbaas-healthmonitor-update': lbaas_healthmon.UpdateHealthMonitor,
+    'lbaas-healthmonitor-delete': lbaas_healthmon.DeleteHealthMonitor,
+    'lbaas-member-list': lbaas_member.ListMember,
+    'lbaas-member-show': lbaas_member.ShowMember,
+    'lbaas-member-create': lbaas_member.CreateMember,
+    'lbaas-member-update': lbaas_member.UpdateMember,
+    'lbaas-member-delete': lbaas_member.DeleteMember,
     'lb-vip-list': lb_vip.ListVip,
     'lb-vip-show': lb_vip.ShowVip,
     'lb-vip-create': lb_vip.CreateVip,
@@ -200,6 +283,10 @@ COMMAND_V2 = {
     'l3-agent-list-hosting-router': agentscheduler.ListL3AgentsHostingRouter,
     'lb-pool-list-on-agent': agentscheduler.ListPoolsOnLbaasAgent,
     'lb-agent-hosting-pool': agentscheduler.GetLbaasAgentHostingPool,
+    'lbaas-loadbalancer-list-on-agent':
+        agentscheduler.ListLoadBalancersOnLbaasAgent,
+    'lbaas-agent-hosting-loadbalancer':
+        agentscheduler.GetLbaasAgentHostingLoadBalancer,
     'service-provider-list': servicetype.ListServiceProvider,
     'firewall-rule-list': firewallrule.ListFirewallRule,
     'firewall-rule-show': firewallrule.ShowFirewallRule,
@@ -272,11 +359,37 @@ COMMAND_V2 = {
     'nuage-netpartition-show': netpartition.ShowNetPartition,
     'nuage-netpartition-create': netpartition.CreateNetPartition,
     'nuage-netpartition-delete': netpartition.DeleteNetPartition,
-    'nec-packet-filter-list': packetfilter.ListPacketFilter,
-    'nec-packet-filter-show': packetfilter.ShowPacketFilter,
-    'nec-packet-filter-create': packetfilter.CreatePacketFilter,
-    'nec-packet-filter-update': packetfilter.UpdatePacketFilter,
-    'nec-packet-filter-delete': packetfilter.DeletePacketFilter,
+    'rbac-create': rbac.CreateRBACPolicy,
+    'rbac-update': rbac.UpdateRBACPolicy,
+    'rbac-list': rbac.ListRBACPolicy,
+    'rbac-show': rbac.ShowRBACPolicy,
+    'rbac-delete': rbac.DeleteRBACPolicy,
+    'address-scope-list': address_scope.ListAddressScope,
+    'address-scope-show': address_scope.ShowAddressScope,
+    'address-scope-create': address_scope.CreateAddressScope,
+    'address-scope-delete': address_scope.DeleteAddressScope,
+    'address-scope-update': address_scope.UpdateAddressScope,
+    'qos-policy-list': qos_policy.ListQoSPolicy,
+    'qos-policy-show': qos_policy.ShowQoSPolicy,
+    'qos-policy-create': qos_policy.CreateQoSPolicy,
+    'qos-policy-update': qos_policy.UpdateQoSPolicy,
+    'qos-policy-delete': qos_policy.DeleteQoSPolicy,
+    'qos-bandwidth-limit-rule-create': (
+        bandwidth_limit_rule.CreateQoSBandwidthLimitRule
+    ),
+    'qos-bandwidth-limit-rule-show': (
+        bandwidth_limit_rule.ShowQoSBandwidthLimitRule
+    ),
+    'qos-bandwidth-limit-rule-list': (
+        bandwidth_limit_rule.ListQoSBandwidthLimitRules
+    ),
+    'qos-bandwidth-limit-rule-update': (
+        bandwidth_limit_rule.UpdateQoSBandwidthLimitRule
+    ),
+    'qos-bandwidth-limit-rule-delete': (
+        bandwidth_limit_rule.DeleteQoSBandwidthLimitRule
+    ),
+    'qos-available-rule-types': qos_rule.ListQoSRuleTypes,
 }
 
 COMMANDS = {'2.0': COMMAND_V2}
@@ -326,6 +439,11 @@ class NeutronShell(app.App):
         for k, v in self.commands[apiversion].items():
             self.command_manager.add_command(k, v)
 
+        self._register_extensions(VERSION)
+
+        # Pop the 'complete' to correct the outputs of 'neutron help'.
+        self.command_manager.commands.pop('complete')
+
         # This is instantiated in initialize_app() only when using
         # password flow auth
         self.auth_client = None
@@ -368,13 +486,57 @@ class NeutronShell(app.App):
             nargs=0,
             default=self,  # tricky
             help=_("Show this help message and exit."))
-        # Global arguments
+        parser.add_argument(
+            '-r', '--retries',
+            metavar="NUM",
+            type=check_non_negative_int,
+            default=0,
+            help=_("How many times the request to the Neutron server should "
+                   "be retried if it fails."))
+        # FIXME(bklei): this method should come from python-keystoneclient
+        self._append_global_identity_args(parser)
+
+        return parser
+
+    def _append_global_identity_args(self, parser):
+        # FIXME(bklei): these are global identity (Keystone) arguments which
+        # should be consistent and shared by all service clients. Therefore,
+        # they should be provided by python-keystoneclient. We will need to
+        # refactor this code once this functionality is available in
+        # python-keystoneclient.
+        #
+        # Note: At that time we'll need to decide if we can just abandon
+        #       the deprecated args (--service-type and --endpoint-type).
+
+        parser.add_argument(
+            '--os-service-type', metavar='<os-service-type>',
+            default=env('OS_NETWORK_SERVICE_TYPE', default='network'),
+            help=_('Defaults to env[OS_NETWORK_SERVICE_TYPE] or network.'))
+
+        parser.add_argument(
+            '--os-endpoint-type', metavar='<os-endpoint-type>',
+            default=env('OS_ENDPOINT_TYPE', default='publicURL'),
+            help=_('Defaults to env[OS_ENDPOINT_TYPE] or publicURL.'))
+
+        # FIXME(bklei): --service-type is deprecated but kept in for
+        # backward compatibility.
+        parser.add_argument(
+            '--service-type', metavar='<service-type>',
+            default=env('OS_NETWORK_SERVICE_TYPE', default='network'),
+            help=_('DEPRECATED! Use --os-service-type.'))
+
+        # FIXME(bklei): --endpoint-type is deprecated but kept in for
+        # backward compatibility.
+        parser.add_argument(
+            '--endpoint-type', metavar='<endpoint-type>',
+            default=env('OS_ENDPOINT_TYPE', default='publicURL'),
+            help=_('DEPRECATED! Use --os-endpoint-type.'))
+
         parser.add_argument(
             '--os-auth-strategy', metavar='<auth-strategy>',
             default=env('OS_AUTH_STRATEGY', default='keystone'),
-            help=_('Authentication strategy, defaults to '
-            'env[OS_AUTH_STRATEGY] or keystone. For now, any '
-            'other value will disable the authentication.'))
+            help=_('DEPRECATED! Only keystone is supported.'))
+
         parser.add_argument(
             '--os_auth_strategy',
             help=argparse.SUPPRESS)
@@ -387,20 +549,39 @@ class NeutronShell(app.App):
             '--os_auth_url',
             help=argparse.SUPPRESS)
 
-        parser.add_argument(
+        project_name_group = parser.add_mutually_exclusive_group()
+        project_name_group.add_argument(
             '--os-tenant-name', metavar='<auth-tenant-name>',
             default=env('OS_TENANT_NAME'),
             help=_('Authentication tenant name, defaults to '
                    'env[OS_TENANT_NAME].'))
+        project_name_group.add_argument(
+            '--os-project-name',
+            metavar='<auth-project-name>',
+            default=utils.env('OS_PROJECT_NAME'),
+            help='Another way to specify tenant name. '
+                 'This option is mutually exclusive with '
+                 ' --os-tenant-name. '
+                 'Defaults to env[OS_PROJECT_NAME].')
+
         parser.add_argument(
             '--os_tenant_name',
             help=argparse.SUPPRESS)
 
-        parser.add_argument(
+        project_id_group = parser.add_mutually_exclusive_group()
+        project_id_group.add_argument(
             '--os-tenant-id', metavar='<auth-tenant-id>',
             default=env('OS_TENANT_ID'),
             help=_('Authentication tenant ID, defaults to '
                    'env[OS_TENANT_ID].'))
+        project_id_group.add_argument(
+            '--os-project-id',
+            metavar='<auth-project-id>',
+            default=utils.env('OS_PROJECT_ID'),
+            help='Another way to specify tenant ID. '
+            'This option is mutually exclusive with '
+            ' --os-tenant-id. '
+            'Defaults to env[OS_PROJECT_ID].')
 
         parser.add_argument(
             '--os-username', metavar='<auth-username>',
@@ -415,6 +596,78 @@ class NeutronShell(app.App):
             default=env('OS_USER_ID'),
             help=_('Authentication user ID (Env: OS_USER_ID)'))
 
+        parser.add_argument(
+            '--os_user_id',
+            help=argparse.SUPPRESS)
+
+        parser.add_argument(
+            '--os-user-domain-id',
+            metavar='<auth-user-domain-id>',
+            default=utils.env('OS_USER_DOMAIN_ID'),
+            help='OpenStack user domain ID. '
+            'Defaults to env[OS_USER_DOMAIN_ID].')
+
+        parser.add_argument(
+            '--os_user_domain_id',
+            help=argparse.SUPPRESS)
+
+        parser.add_argument(
+            '--os-user-domain-name',
+            metavar='<auth-user-domain-name>',
+            default=utils.env('OS_USER_DOMAIN_NAME'),
+            help='OpenStack user domain name. '
+                 'Defaults to env[OS_USER_DOMAIN_NAME].')
+
+        parser.add_argument(
+            '--os_user_domain_name',
+            help=argparse.SUPPRESS)
+
+        parser.add_argument(
+            '--os_project_id',
+            help=argparse.SUPPRESS)
+
+        parser.add_argument(
+            '--os_project_name',
+            help=argparse.SUPPRESS)
+
+        parser.add_argument(
+            '--os-project-domain-id',
+            metavar='<auth-project-domain-id>',
+            default=utils.env('OS_PROJECT_DOMAIN_ID'),
+            help='Defaults to env[OS_PROJECT_DOMAIN_ID].')
+
+        parser.add_argument(
+            '--os-project-domain-name',
+            metavar='<auth-project-domain-name>',
+            default=utils.env('OS_PROJECT_DOMAIN_NAME'),
+            help='Defaults to env[OS_PROJECT_DOMAIN_NAME].')
+
+        parser.add_argument(
+            '--os-cert',
+            metavar='<certificate>',
+            default=utils.env('OS_CERT'),
+            help=_("Path of certificate file to use in SSL "
+                   "connection. This file can optionally be "
+                   "prepended with the private key. Defaults "
+                   "to env[OS_CERT]."))
+
+        parser.add_argument(
+            '--os-cacert',
+            metavar='<ca-certificate>',
+            default=env('OS_CACERT', default=None),
+            help=_("Specify a CA bundle file to use in "
+                   "verifying a TLS (https) server certificate. "
+                   "Defaults to env[OS_CACERT]."))
+
+        parser.add_argument(
+            '--os-key',
+            metavar='<key>',
+            default=utils.env('OS_KEY'),
+            help=_("Path of client key to use in SSL "
+                   "connection. This option is not necessary "
+                   "if your key is prepended to your certificate "
+                   "file. Defaults to env[OS_KEY]."))
+
         parser.add_argument(
             '--os-password', metavar='<auth-password>',
             default=utils.env('OS_PASSWORD'),
@@ -441,21 +694,11 @@ class NeutronShell(app.App):
             help=argparse.SUPPRESS)
 
         parser.add_argument(
-            '--service-type', metavar='<service-type>',
-            default=env('OS_NETWORK_SERVICE_TYPE', default='network'),
-            help=_('Defaults to env[OS_NETWORK_SERVICE_TYPE] or network.'))
-
-        parser.add_argument(
-            '--timeout', metavar='<seconds>',
+            '--http-timeout', metavar='<seconds>',
             default=env('OS_NETWORK_TIMEOUT', default=None), type=float,
             help=_('Timeout in seconds to wait for an HTTP response. Defaults '
                    'to env[OS_NETWORK_TIMEOUT] or None if not specified.'))
 
-        parser.add_argument(
-            '--endpoint-type', metavar='<endpoint-type>',
-            default=env('OS_ENDPOINT_TYPE', default='publicURL'),
-            help=_('Defaults to env[OS_ENDPOINT_TYPE] or publicURL.'))
-
         parser.add_argument(
             '--os-url', metavar='<url>',
             default=env('OS_URL'),
@@ -464,14 +707,6 @@ class NeutronShell(app.App):
             '--os_url',
             help=argparse.SUPPRESS)
 
-        parser.add_argument(
-            '--os-cacert',
-            metavar='<ca-certificate>',
-            default=env('OS_CACERT', default=None),
-            help=_("Specify a CA bundle file to use in "
-                   "verifying a TLS (https) server certificate. "
-                   "Defaults to env[OS_CACERT]."))
-
         parser.add_argument(
             '--insecure',
             action='store_true',
@@ -481,8 +716,6 @@ class NeutronShell(app.App):
                    "not be verified against any certificate authorities. "
                    "This option should be used with caution."))
 
-        return parser
-
     def _bash_completion(self):
         """Prints all of the commands and options for bash-completion."""
         commands = set()
@@ -498,6 +731,29 @@ class NeutronShell(app.App):
                 options.add(option)
         print(' '.join(commands | options))
 
+    def _register_extensions(self, version):
+        for name, module in itertools.chain(
+                client_extension._discover_via_entry_points()):
+            self._extend_shell_commands(name, module, version)
+
+    def _extend_shell_commands(self, name, module, version):
+        classes = inspect.getmembers(module, inspect.isclass)
+        for cls_name, cls in classes:
+            if (issubclass(cls, client_extension.NeutronClientExtension) and
+                    hasattr(cls, 'shell_command')):
+                cmd = cls.shell_command
+                if hasattr(cls, 'versions'):
+                    if version not in cls.versions:
+                        continue
+                try:
+                    name_prefix = "[%s]" % name
+                    cls.__doc__ = ("%s %s" % (name_prefix, cls.__doc__) if
+                                   cls.__doc__ else name_prefix)
+                    self.command_manager.add_command(cmd, cls)
+                    self.commands[version][cmd] = cls
+                except TypeError:
+                    pass
+
     def run(self, argv):
         """Equivalent to the main program for the application.
 
@@ -510,7 +766,7 @@ class NeutronShell(app.App):
             help_pos = -1
             help_command_pos = -1
             for arg in argv:
-                if arg == 'bash-completion':
+                if arg == 'bash-completion' and help_command_pos == -1:
                     self._bash_completion()
                     return 0
                 if arg in self.commands[self.api_version]:
@@ -532,27 +788,22 @@ class NeutronShell(app.App):
             self.interactive_mode = not remainder
             self.initialize_app(remainder)
         except Exception as err:
-            if self.options.verbose_level == self.DEBUG_LEVEL:
-                self.log.exception(unicode(err))
+            if self.options.verbose_level >= self.DEBUG_LEVEL:
+                self.log.exception(err)
                 raise
             else:
-                self.log.error(unicode(err))
+                self.log.error(err)
             return 1
-        result = 1
         if self.interactive_mode:
             _argv = [sys.argv[0]]
             sys.argv = _argv
-            result = self.interact()
-        else:
-            result = self.run_subcommand(remainder)
-        return result
+            return self.interact()
+        return self.run_subcommand(remainder)
 
     def run_subcommand(self, argv):
         subcommand = self.command_manager.find_command(argv)
         cmd_factory, cmd_name, sub_argv = subcommand
         cmd = cmd_factory(self, self.options)
-        err = None
-        result = 1
         try:
             self.prepare_to_run_command(cmd)
             full_name = (cmd_name
@@ -561,29 +812,12 @@ class NeutronShell(app.App):
                          )
             cmd_parser = cmd.get_parser(full_name)
             return run_command(cmd, cmd_parser, sub_argv)
-        except Exception as err:
-            if self.options.verbose_level == self.DEBUG_LEVEL:
-                self.log.exception(unicode(err))
-            else:
-                self.log.error(unicode(err))
-            try:
-                self.clean_up(cmd, result, err)
-            except Exception as err2:
-                if self.options.verbose_level == self.DEBUG_LEVEL:
-                    self.log.exception(unicode(err2))
-                else:
-                    self.log.error(_('Could not clean up: %s'), unicode(err2))
-            if self.options.verbose_level == self.DEBUG_LEVEL:
+        except Exception as e:
+            if self.options.verbose_level >= self.DEBUG_LEVEL:
+                self.log.exception("%s", e)
                 raise
-        else:
-            try:
-                self.clean_up(cmd, result, None)
-            except Exception as err3:
-                if self.options.verbose_level == self.DEBUG_LEVEL:
-                    self.log.exception(unicode(err3))
-                else:
-                    self.log.error(_('Could not clean up: %s'), unicode(err3))
-        return result
+            self.log.error("%s", e)
+        return 1
 
     def authenticate_user(self):
         """Make sure the user has provided all of the authentication
@@ -595,43 +829,74 @@ class NeutronShell(app.App):
                 if not self.options.os_token:
                     raise exc.CommandError(
                         _("You must provide a token via"
-                          " either --os-token or env[OS_TOKEN]"))
+                          " either --os-token or env[OS_TOKEN]"
+                          " when providing a service URL"))
 
                 if not self.options.os_url:
                     raise exc.CommandError(
                         _("You must provide a service URL via"
-                          " either --os-url or env[OS_URL]"))
+                          " either --os-url or env[OS_URL]"
+                          " when providing a token"))
 
             else:
                 # Validate password flow auth
+                project_info = (self.options.os_tenant_name or
+                                self.options.os_tenant_id or
+                                (self.options.os_project_name and
+                                    (self.options.os_project_domain_name or
+                                     self.options.os_project_domain_id)) or
+                                self.options.os_project_id)
+
                 if (not self.options.os_username
-                    and not self.options.os_user_id):
+                        and not self.options.os_user_id):
                     raise exc.CommandError(
                         _("You must provide a username or user ID via"
                           "  --os-username, env[OS_USERNAME] or"
-                          "  --os-user_id, env[OS_USER_ID]"))
+                          "  --os-user-id, env[OS_USER_ID]"))
 
                 if not self.options.os_password:
-                    raise exc.CommandError(
-                        _("You must provide a password via"
-                          " either --os-password or env[OS_PASSWORD]"))
+                    # No password, If we've got a tty, try prompting for it
+                    if hasattr(sys.stdin, 'isatty') and sys.stdin.isatty():
+                        # Check for Ctl-D
+                        try:
+                            self.options.os_password = getpass.getpass(
+                                'OS Password: ')
+                        except EOFError:
+                            pass
+                    # No password because we didn't have a tty or the
+                    # user Ctl-D when prompted.
+                    if not self.options.os_password:
+                        raise exc.CommandError(
+                            _("You must provide a password via"
+                              " either --os-password or env[OS_PASSWORD]"))
 
-                if (not self.options.os_tenant_name
-                    and not self.options.os_tenant_id):
+                if (not project_info):
+                    # tenent is deprecated in Keystone v3. Use the latest
+                    # terminology instead.
                     raise exc.CommandError(
-                        _("You must provide a tenant_name or tenant_id via"
-                          "  --os-tenant-name, env[OS_TENANT_NAME]"
-                          "  --os-tenant-id, or via env[OS_TENANT_ID]"))
+                        _("You must provide a project_id or project_name ("
+                          "with project_domain_name or project_domain_id) "
+                          "via "
+                          "  --os-project-id (env[OS_PROJECT_ID])"
+                          "  --os-project-name (env[OS_PROJECT_NAME]),"
+                          "  --os-project-domain-id "
+                          "(env[OS_PROJECT_DOMAIN_ID])"
+                          "  --os-project-domain-name "
+                          "(env[OS_PROJECT_DOMAIN_NAME])"))
 
                 if not self.options.os_auth_url:
                     raise exc.CommandError(
                         _("You must provide an auth url via"
                           " either --os-auth-url or via env[OS_AUTH_URL]"))
+            auth_session = self._get_keystone_session()
+            auth = auth_session.auth
         else:   # not keystone
             if not self.options.os_url:
                 raise exc.CommandError(
                     _("You must provide a service URL via"
                       " either --os-url or env[OS_URL]"))
+            auth_session = None
+            auth = None
 
         self.client_manager = clientmanager.ClientManager(
             token=self.options.os_token,
@@ -645,11 +910,18 @@ class NeutronShell(app.App):
             region_name=self.options.os_region_name,
             api_version=self.api_version,
             auth_strategy=self.options.os_auth_strategy,
-            service_type=self.options.service_type,
-            endpoint_type=self.options.endpoint_type,
+            # FIXME (bklei) honor deprecated service_type and
+            # endpoint type until they are removed
+            service_type=self.options.os_service_type or
+            self.options.service_type,
+            endpoint_type=self.options.os_endpoint_type or self.endpoint_type,
             insecure=self.options.insecure,
             ca_cert=self.options.os_cacert,
-            timeout=self.options.timeout,
+            timeout=self.options.http_timeout,
+            retries=self.options.retries,
+            raise_errors=False,
+            session=auth_session,
+            auth=auth,
             log_credentials=True)
         return
 
@@ -673,11 +945,6 @@ class NeutronShell(app.App):
         if self.interactive_mode or cmd_name != 'help':
             self.authenticate_user()
 
-    def clean_up(self, cmd, result, err):
-        self.log.debug('clean_up %s', cmd.__class__.__name__)
-        if err:
-            self.log.debug('Got an error: %s', unicode(err))
-
     def configure_logging(self):
         """Create logging handlers for any log output."""
         root_logger = logging.getLogger('')
@@ -691,25 +958,118 @@ class NeutronShell(app.App):
                          self.INFO_LEVEL: logging.INFO,
                          self.DEBUG_LEVEL: logging.DEBUG,
                          }.get(self.options.verbose_level, logging.DEBUG)
-        console.setLevel(console_level)
+        # The default log level is INFO, in this situation, set the
+        # log level of the console to WARNING, to avoid displaying
+        # useless messages. This equals using "--quiet"
+        if console_level == logging.INFO:
+            console.setLevel(logging.WARNING)
+        else:
+            console.setLevel(console_level)
         if logging.DEBUG == console_level:
             formatter = logging.Formatter(self.DEBUG_MESSAGE_FORMAT)
         else:
             formatter = logging.Formatter(self.CONSOLE_MESSAGE_FORMAT)
+        logging.getLogger('iso8601.iso8601').setLevel(logging.WARNING)
         logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
         console.setFormatter(formatter)
         root_logger.addHandler(console)
         return
 
+    def get_v2_auth(self, v2_auth_url):
+        return v2_auth.Password(
+            v2_auth_url,
+            username=self.options.os_username,
+            password=self.options.os_password,
+            tenant_id=self.options.os_tenant_id,
+            tenant_name=self.options.os_tenant_name)
+
+    def get_v3_auth(self, v3_auth_url):
+        project_id = self.options.os_project_id or self.options.os_tenant_id
+        project_name = (self.options.os_project_name or
+                        self.options.os_tenant_name)
+
+        return v3_auth.Password(
+            v3_auth_url,
+            username=self.options.os_username,
+            password=self.options.os_password,
+            user_id=self.options.os_user_id,
+            user_domain_name=self.options.os_user_domain_name,
+            user_domain_id=self.options.os_user_domain_id,
+            project_id=project_id,
+            project_name=project_name,
+            project_domain_name=self.options.os_project_domain_name,
+            project_domain_id=self.options.os_project_domain_id
+        )
+
+    def _discover_auth_versions(self, session, auth_url):
+        # discover the API versions the server is supporting base on the
+        # given URL
+        try:
+            ks_discover = discover.Discover(session=session, auth_url=auth_url)
+            return (ks_discover.url_for('2.0'), ks_discover.url_for('3.0'))
+        except ks_exc.ClientException:
+            # Identity service may not support discover API version.
+            # Lets try to figure out the API version from the original URL.
+            url_parts = urlparse.urlparse(auth_url)
+            (scheme, netloc, path, params, query, fragment) = url_parts
+            path = path.lower()
+            if path.startswith('/v3'):
+                return (None, auth_url)
+            elif path.startswith('/v2'):
+                return (auth_url, None)
+            else:
+                # not enough information to determine the auth version
+                msg = _('Unable to determine the Keystone version '
+                        'to authenticate with using the given '
+                        'auth_url. Identity service may not support API '
+                        'version discovery. Please provide a versioned '
+                        'auth_url instead.')
+                raise exc.CommandError(msg)
+
+    def _get_keystone_session(self):
+        # first create a Keystone session
+        cacert = self.options.os_cacert or None
+        cert = self.options.os_cert or None
+        key = self.options.os_key or None
+        insecure = self.options.insecure or False
+        ks_session = session.Session.construct(dict(cacert=cacert,
+                                                    cert=cert,
+                                                    key=key,
+                                                    insecure=insecure))
+        # discover the supported keystone versions using the given url
+        (v2_auth_url, v3_auth_url) = self._discover_auth_versions(
+            session=ks_session,
+            auth_url=self.options.os_auth_url)
+
+        # Determine which authentication plugin to use. First inspect the
+        # auth_url to see the supported version. If both v3 and v2 are
+        # supported, then use the highest version if possible.
+        user_domain_name = self.options.os_user_domain_name or None
+        user_domain_id = self.options.os_user_domain_id or None
+        project_domain_name = self.options.os_project_domain_name or None
+        project_domain_id = self.options.os_project_domain_id or None
+        domain_info = (user_domain_name or user_domain_id or
+                       project_domain_name or project_domain_id)
+
+        if (v2_auth_url and not domain_info) or not v3_auth_url:
+            ks_session.auth = self.get_v2_auth(v2_auth_url)
+        else:
+            ks_session.auth = self.get_v3_auth(v3_auth_url)
+
+        return ks_session
+
 
 def main(argv=sys.argv[1:]):
     try:
-        return NeutronShell(NEUTRON_API_VERSION).run(map(strutils.safe_decode,
-                                                         argv))
+        return NeutronShell(NEUTRON_API_VERSION).run(
+            list(map(encodeutils.safe_decode, argv)))
+    except KeyboardInterrupt:
+        print("... terminating neutron client", file=sys.stderr)
+        return 130
     except exc.NeutronClientException:
         return 1
     except Exception as e:
-        print(unicode(e))
+        print(e)
         return 1
 
 
